UNIT ComboBox;
{+H
---------------------------------------------------------------------------
  File        - COMBOBOX.PAS
  Purpose     - This component completes text that user writes in combobox
                with items presents on the list. Like Internet Explorer 4 does.
  Version     - TComboBox98 component version 1.0 (12/12/97)
  Usage       - Component for Delphi 1.0, 2.0, 3.0x and C++ Builder 1.0.
  Author      - CRESTO Sylvain
                email: cresto@mygale.org
                http://www.mygale.org/~cresto/
                La Borde / 09230 / Contrazy / France
  Copyright    1997 CRESTO Sylvain
  Restrictions- This component is FREEWARE!
                No warranty is given by the author, expressed or implied.
                Do not distribute modified version without my agreement. thanks.
  Revised     - 1998.0227 (KSB) Reformatted & renamed variables, etc. to standard format.
              - 1998.0302 (KSB) Modified to add Quicken-like behavior when selected.
              - 1998.0303 (KSB) Added proper behavior when <ENTER> is pressed.
---------------------------------------------------------------------------}
INTERFACE
USES
{$IfDef Win32}
  Windows,
{$Else}
  WinTypes, WinProcs,
{$EndIf}
  Messages, SysUtils, Classes,  Graphics, Controls, Forms,    Dialogs,
  StdCtrls, DsgnIntf;

TYPE
  TOnComplete = PROCEDURE(aSender:TObject; aKey:WORD; VAR aCompleteWith:STRING; VAR aDoComplete:BOOLEAN) OF OBJECT;
  TAboutProperty   = CLASS(TPropertyEditor)
  PRIVATE
    PROCEDURE   Edit; OVERRIDE;
    FUNCTION    GetAttributes: TPropertyAttributes; OVERRIDE;
    FUNCTION    GetValue: STRING; OVERRIDE;
  END {CLASS};

  TComboBox98 = CLASS(TComboBox)
  PRIVATE
    FfComplete: BOOLEAN;
    FwKey     : WORD;
    FwMinLngth: WORD;
    FsText    : STRING;
    { .. }
    FAbout    : TAboutProperty;
    FOnComplete:TOnComplete;
    PROCEDURE   SetComplete(aValue:BOOLEAN);
    PROCEDURE   SetMinLength(aValue:WORD);
  PUBLIC
    CONSTRUCTOR Create(aOwner : TComponent); OVERRIDE;
  PROTECTED
    PROCEDURE   Change; OVERRIDE;
    PROCEDURE   DoEnter; OVERRIDE;
    PROCEDURE   KeyDown(VAR aKey:WORD; aShift:TShiftState); OVERRIDE;
    PROPERTY    StoredText : STRING READ FsText WRITE FsText;
  PUBLISHED
    PROPERTY    About : TAboutProperty READ FAbout WRITE FAbout;
    PROPERTY    Complete : BOOLEAN READ FfComplete WRITE SetComplete;
    PROPERTY    MinLengthComplete : WORD READ FwMinLngth WRITE SetMinLength;
    PROPERTY    OnComplete : TOnComplete READ FOnComplete WRITE FOnComplete;
  END {CLASS};

              {=================================================}
IMPLEMENTATION

{$IfDef Win32}
{$R COMBOBOX98.RES}
{$Else}
{}FUNCTION  MakeLParam(awLow, awHigh: WORD): LongINT;
{+H
---------------------------------------------------------------------------
  Purpose     - 16 bit substitution for the MAKELPARAM macro.
                The MAKELPARAM macro creates an unsigned 32-bit value for
                use as an lParam parameter in a message. The macro
                concatenates two given 16-bit values.

  Parameters  -
        wLow    Specifies the low-order word of the new long value.
        wHigh   Specifies the high-order word of the new long value.

  Return Value- The return value is an unsigned 32-bit value.

  Remarks     - The MAKELPARAM macro is defined as follows:
                #define MAKELPARAM(l, h)   ((LPARAM) MAKELONG(l, h))
---------------------------------------------------------------------------}
  BEGIN
    Result := awLow OR awHigh SHL 16;
{}END {MakeLParam};
{$R COMBOB16.RES}
{$EndIf}

(*
procedure Register;
begin
  RegisterComponents('Standard', [TComboBox98]);
  RegisterPropertyEditor(TypeInfo(TAboutProperty), TComboBox98, 'ABOUT', TAboutProperty);
END;
*)

{ TComboBox98 }

{}CONSTRUCTOR TComboBox98.Create(aOwner : Tcomponent);
  BEGIN
    INHERITED Create(aOwner);
    FfComplete := TRUE;
    FwKey := 0;
    FwMinLngth := 0;
    FsText := '';
{}END {Create};




{}PROCEDURE   TComboBox98.Change;
{+M
---------------------------------------------------------------------------
  Revised     - 1998.0227 (KSB) Reformatted code. Revised variable names.
              - 1998.0302 (KSB) Replaced the magic numbers 8 and 46 with the
                virtual key codes VK_BACK and VK_DELETE respectively. Revised
                code to reselect nearest match when deleting chars.
---------------------------------------------------------------------------}
  VAR
    lfD       : BOOLEAN;
    lfDoIt    : BOOLEAN;
    lw        : WORD;
    lwD       : WORD;
    ls        : STRING;
  BEGIN
    INHERITED Change;
    IF (FfComplete) AND ((Style=csDropDown) OR (Style=csSimple)) THEN BEGIN
      IF (Length(Text)>=FwMinLngth) AND (Items.Count > 0) THEN BEGIN
        IF (FwKey<>VK_TAB) AND (FwKey<>VK_ESCAPE) THEN BEGIN {don't trigger on a "TAB" or "ESCAPE" key}
          ls := UpperCase(Text);            {case insensitive search}
          lfD:= (FwKey=VK_BACK) OR (FwKey=VK_DELETE);

          FOR lw := 0 TO items.Count-1 DO BEGIN
            IF ls = UpperCase(Copy(items.strings[lw],1,Length(ls))) THEN BEGIN
              IF lfD THEN BEGIN             {"BACKSPACE" or "DELETE" key}
                ls := Text;
                ItemIndex := lw;
                Text := ls;
                SendMessage(Handle, CB_SETEDITSEL, 0, MakeLParam(Length(Text),Length(Text)));
              END ELSE BEGIN                {all keys, not [BACKSPACE,DELETE,TAB,ESCAPE]}
                lwD := Length(ls);
                ls := items.strings[lw];
                ItemIndex := lw;
                lfDoIt := TRUE;
                IF Assigned(FOnComplete) THEN BEGIN
                  FOnComplete(Self,FwKey,ls,lfDoIt);
                END {IF};

                IF lfDoIt THEN BEGIN
                  Text := ls;
                  SendMessage(Handle, CB_SETEDITSEL, 0, MakeLParam(lwD,Length(Text)));
                END {IF};
              END {IF};

              Break;
            END {IF};
          END {FOR};
        END {IF};
      END {IF};
    END {IF};
{}END {Change};




{}PROCEDURE   TComboBox98.DoEnter;
{+M
---------------------------------------------------------------------------
  Purpose     - Override the default DoEnter method.
  Remarks     - This method forces the following behaviors:
                1. the original text is saved and restored. Otherwise it is
                   cleared when the list box drops down. 
                2. the list box will appear automatically when the control
                   receives the focus.
                3. It also resets the itemIndex to "unselected" (-1) [if
                   the control was visited before, it remembers the previous
                   selection from the list (default is 0) and forces the text
                   to contain that value unless the index is reset]
---------------------------------------------------------------------------}
  BEGIN
    FsText := Text;
    INHERITED DoEnter;
    DroppedDown := TRUE;
    ItemIndex := -1;
    Text := FsText;
{}END {DoEnter};




{}PROCEDURE   TComboBox98.KeyDown(VAR aKey: WORD; aShift: TShiftState);
{+M
---------------------------------------------------------------------------
  Purpose     - Override the default KeyDown method.
  Remarks     - This method forces the following behaviors:
                1. if the key is VK_TAB the list box is closed.
                2. if the key is VK_ESCAPE the list box is closed and the
                   original text value (the text value before the combo was
                   selected) is restored.
                3. if the key is VK_RETURN the closest match to the current
                   text value is selected if text is a proper substring of
                   any string in the list.
                4. For any other key, the list box is opened.
---------------------------------------------------------------------------}
  VAR
    ls        : STRING;
  BEGIN
    ls := Text;
    INHERITED KeyDown(aKey,aShift);
    FwKey := aKey;
    CASE FwKey OF
      VK_TAB    : DroppedDown := FALSE;
      VK_ESCAPE : BEGIN
        ItemIndex := -1;
        DroppedDown := FALSE;
        Text := FsText;
        SendMessage(Handle, CB_SETEDITSEL, 0, MakeLParam(Length(Text),Length(Text)));
        Exit;
      END;

      VK_RETURN : BEGIN
        ls := Text;
        DroppedDown := FALSE;
        Text := ls;
        Change;
        Exit;
      END;

      ELSE BEGIN
        DroppedDown := TRUE;
      END;
    END {CASE};

    Text := ls;
{}END {KeyDown};




{}PROCEDURE   TComboBox98.SetComplete(aValue:BOOLEAN);
  BEGIN
    FfComplete := aValue;
{}END {SetComplete};




{}PROCEDURE   TComboBox98.SetMinLength(aValue:WORD);
  BEGIN
    FwMinLngth := aValue;
{}END {SetMinLength};



{ TAboutProperty }

{}PROCEDURE   TAboutProperty.Edit;
  BEGIN
    Application.MessageBox
      ('By CRESTO Sylvain - 1997'#13
      +'This component (for Delphi 1.0, 2.0, 3.0x and C++ Builder 1.0) is FREEWARE.'#13#13
      +'E-Mail : cresto@mygale.org'#13
      +'WEB : http://www.mygale.org/~cresto/'#13
      +'Revised by Keith Brown - 1998; Added Quicken-like behavior'#13
      +'E-Mail: keith_brown@juno.com'
      ,'ComboBox98 component version Q1.0',MB_OK+ MB_ICONINFORMATION);
{}END {Edit};




{}FUNCTION    TAboutProperty.GetAttributes: TPropertyAttributes;
  BEGIN
    GetAttributes:=[paDialog, paReadOnly];
{}END {GetAttributes};




{}FUNCTION    TAboutProperty.GetValue: STRING;
  BEGIN
    GetValue:='(About)';
{}END {GetValue};


END {ComboBox}.
