UNIT KNavigat;
{+H
---------------------------------------------------------------------------
  File        - KNAVIGAT.PAS
  Purpose     - Provide a navigator that does not require a TDataSource.
  Author      - Keith S. Brown
  Reference   - DBCtrls.PAS by Borland Int.
  Revised     - 1998.0212 (KSB) Wrote initial version based on TDBNavigator.
              - 1998.0219 (KSB) The dataSet property now has a working property editor (see KToolReg.pas). The
                Glyphs property works in a manner similar to the images in a TTreeView... glyphs are taken from
                an associated TImageList on the form.
---------------------------------------------------------------------------}
{$R-}
INTERFACE
USES
  Windows,  SysUtils, Messages, Classes,  Controls, ExtCtrls, Buttons,  Db,
  DBTables, Forms,    Graphics,
  CommCtrl;


TYPE
  TKNavButton = CLASS;
  TKNavBtn    = (nbFirst, nbPrior,  nbNext, nbLast,
                 nbInsert,nbDelete, nbEdit, nbPost,
                 nbCancel,nbRefresh,nbUser1,nbUser2);
  TKNavBtnSet = SET OF TKNavBtn;
  TKNavBtnStyle    = SET OF(nsAllowTimer,nsFocusRect);
  TKNavBtnClick    = PROCEDURE(Sender: TObject; Button:TKNavBtn)OF OBJECT;

  { TDBNavigator }

  TKNavigator = CLASS(TCustomPanel)
  PRIVATE
    FButtonWidth   : INTEGER;
    FConfirmDelete : BOOLEAN;
    FDataSet       : TDataSet;
    FFlat          : BOOLEAN;
    FFocusedButton : TKNavBtn;
    FGlyphs        : TImageList;
    FHints         : TStrings;
    FMinBtnSize    : TPoint;
    FOnBeforeAction: TKNavBtnClick;
    FOnNavClick    : TKNavBtnClick;
    FVisibleButtons: TKNavBtnSet;

    PROCEDURE   AdjustSize(VAR aW,aH:INTEGER);
    PROCEDURE   BtnMouseDown(aSender:TObject; aButton:TMouseButton; aShift:TShiftState; X,Y:INTEGER);
    PROCEDURE   ClickHandler(aSender:TObject);
    PROCEDURE   CMEnabledChanged(VAR aMessage: TMessage); MESSAGE CM_ENABLEDCHANGED;
    FUNCTION    GetDataSet:TDataSet;
    PROCEDURE   HintsChanged(aSender: TObject);
    PROCEDURE   InitButtons;
    PROCEDURE   InitHints;
    PROCEDURE   SetDataSet(aValue:TDataSet);
    PROCEDURE   SetFlat(aValue: BOOLEAN);
    PROCEDURE   SetHints(aValue: TStrings);
    PROCEDURE   SetVisible(aValue: TKNavBtnSet);
    PROCEDURE   WMGetDlgCode(VAR aMessage: TWMGetDlgCode); MESSAGE WM_GETDLGCODE;
    PROCEDURE   WMKillFocus(VAR aMessage: TWMKillFocus); MESSAGE WM_KILLFOCUS;
    PROCEDURE   WMSetFocus(VAR aMessage: TWMSetFocus); MESSAGE WM_SETFOCUS;
    PROCEDURE   WMSize(VAR aMessage: TWMSize); MESSAGE WM_SIZE;
  PROTECTED
    Buttons   : ARRAY[TKNavBtn]OF TKNavButton;

    PROCEDURE   ActiveChanged;
    PROCEDURE   DataChanged;
    PROCEDURE   DefineProperties(aFiler: TFiler); OVERRIDE;
    PROCEDURE   EditingChanged;
    PROCEDURE   GetChildren(aProc:TGetChildProc; aRoot:TComponent); OVERRIDE;
    PROCEDURE   KeyDown(VAR aKey:WORD; aShift:TShiftState); OVERRIDE;
    PROCEDURE   Loaded; OVERRIDE;
    PROCEDURE   Notification(aComponent:TComponent; aOperation:TOperation); OVERRIDE;
    PROCEDURE   ReadDataSetName(aReader : TReader);
    PROCEDURE   SetImageList(aValue: TImageList);
    FUNCTION    StoreDataSet : Boolean;
    PROCEDURE   WriteDataSetName(aWriter : TWriter);
  PUBLIC
    PROCEDURE   BtnClick(aIndex:TKNavBtn);
    CONSTRUCTOR Create(aOwner:TComponent); OVERRIDE;
    DESTRUCTOR  Destroy; OVERRIDE;
    PROCEDURE   SetBounds(aLeft,aTop,aWidth,aHeight:INTEGER); OVERRIDE;
  PUBLISHED
    PROPERTY    Align;
    PROPERTY    ConfirmDelete: BOOLEAN READ FConfirmDelete WRITE FConfirmDelete DEFAULT TRUE;
    PROPERTY    Ctl3D;
    PROPERTY    DataSet: TDataSet READ GetDataSet WRITE SetDataSet {STORED StoreDataSet};
    PROPERTY    DragCursor;
    PROPERTY    DragMode;
    PROPERTY    Enabled;
    PROPERTY    Flat: BOOLEAN READ FFlat WRITE SetFlat DEFAULT FALSE;
    PROPERTY    Glyphs:TImageList READ FGlyphs WRITE SetImageList;
    PROPERTY    Hints: TStrings READ FHints WRITE SetHints;
    PROPERTY    OnBeforeAction: TKNavBtnClick READ FOnBeforeAction WRITE FOnBeforeAction;
    PROPERTY    OnClick: TKNavBtnClick READ FOnNavClick WRITE FOnNavClick;
    PROPERTY    OnDblClick;
    PROPERTY    OnDragDrop;
    PROPERTY    OnDragOver;
    PROPERTY    OnEndDrag;
    PROPERTY    OnEnter;
    PROPERTY    OnExit;
    PROPERTY    OnResize;
    PROPERTY    OnStartDrag;
    PROPERTY    ParentCtl3D;
    PROPERTY    ParentShowHint;
    PROPERTY    PopupMenu;
    PROPERTY    ShowHint;
    PROPERTY    TabOrder;
    PROPERTY    TabStop;
    PROPERTY    Visible;
    PROPERTY    VisibleButtons: TKNavBtnSet READ FVisibleButtons WRITE SetVisible DEFAULT[nbFirst,nbPrior,nbNext,nbLast,nbInsert,nbDelete,nbEdit,nbPost,nbCancel,nbRefresh];
  END {CLASS};

  { TKNavButton }

  TKNavButton  = CLASS(TSpeedButton)
  PRIVATE
    FIndex    : TKNavBtn;
    FNavStyle : TKNavBtnStyle;
    FRepeatTimer   : TTimer;

    PROCEDURE   TimerExpired(aSender: TObject);
  PROTECTED
    PROCEDURE   MouseDown(aButton:TMouseButton; aShift:TShiftState; X,Y:INTEGER); OVERRIDE;
    PROCEDURE   MouseUp(aButton:TMouseButton; aShift:TShiftState; X,Y:INTEGER); OVERRIDE;
    PROCEDURE   Paint; OVERRIDE;
  PUBLIC
    DESTRUCTOR  Destroy; OVERRIDE;
    PROPERTY    NavStyle: TKNavBtnStyle READ FNavStyle WRITE FNavStyle;
    PROPERTY    Index: TKNavBtn READ FIndex WRITE FIndex;
  END {CLASS};


              {=================================================}
IMPLEMENTATION
USES
  DsgnIntf, Dialogs;

FUNCTION    FindDataSet(CONST aValue:TComponentName):TDataSet; FORWARD;

CONST
  InitRepeatPause  = 400; { pause before repeat timer (ms) }
  RepeatPause = 100;      { pause before hint window displays (ms)}
  SpaceSize   = 5;        { size of space between special buttons }



{}PROCEDURE   TKNavigator.ActiveChanged;
  VAR
    leBtn     : TKNavBtn;
  BEGIN
    IF NOT Enabled THEN BEGIN
      FOR leBtn := Low(Buttons) TO High(Buttons) DO BEGIN
        Buttons[leBtn].Enabled := FALSE
      END {FOR};
    END ELSE BEGIN
      DataChanged;
      EditingChanged;
    END {IF};
{}END {ActiveChanged};




{}PROCEDURE   TKNavigator.AdjustSize(VAR aW,aH:INTEGER);
  VAR
    liCount   : INTEGER;
    liMinW    : INTEGER;
    leBtn     : TKNavBtn;
    liSpace   : INTEGER;
    liTemp    : INTEGER;
    liRemain  : INTEGER;
    liX       : INTEGER;
  BEGIN
    IF (csLoading IN ComponentState) THEN BEGIN
      Exit;
    END {IF};

    IF Buttons[nbFirst] = NIL THEN BEGIN
      Exit;
    END {IF};

    liCount := 0;
    FOR leBtn := Low(Buttons) TO High(Buttons) DO BEGIN
      IF Buttons[leBtn].Visible THEN BEGIN
        Inc(liCount);
      END {IF};
    END {FOR};

    IF liCount = 0 THEN BEGIN
      Inc(liCount);
    END {IF};

    liMinW := liCount*FMinBtnSize.X;
    IF aW<liMinW THEN BEGIN
      aW := liMinW;
    END {IF};

    IF aH<FMinBtnSize.Y THEN BEGIN
      aH := FMinBtnSize.Y;
    END {IF};

    FButtonWidth := aW DIV liCount;
    liTemp := liCount*FButtonWidth;
    IF Align = alNone THEN BEGIN
      aW := liTemp;
    END {IF};

    liX := 0;
    liRemain := aW-liTemp;
    liTemp := liCount DIV 2;
    FOR leBtn := Low(Buttons) TO High(Buttons) DO BEGIN
      IF Buttons[leBtn].Visible THEN BEGIN
        liSpace := 0;
        IF liRemain <> 0 THEN BEGIN
          Dec(liTemp,liRemain);
          IF liTemp<0 THEN BEGIN
            Inc(liTemp,liCount);
            liSpace := 1;
          END {IF};
        END {IF};

        Buttons[leBtn].SetBounds(liX,0,FButtonWidth+liSpace,Height);
        Inc(liX,FButtonWidth+liSpace);
      END ELSE BEGIN
        Buttons[leBtn].SetBounds(Width+1,0,FButtonWidth,Height);
      END {IF};
    END {IF};
{}END {AdjustSize};




{}PROCEDURE   TKNavigator.BtnClick(aIndex:TKNavBtn);
  BEGIN
    IF NOT (csDesigning IN ComponentState) AND Assigned(FOnBeforeAction) THEN BEGIN
      FOnBeforeAction(Self,aIndex);
    END {IF};

    IF (DataSet <> NIL) AND (DataSet.State <> dsInactive) THEN BEGIN
      WITH DataSet DO BEGIN
        CASE aIndex OF
          nbPrior  : Prior;
          nbNext   : Next;
          nbFirst  : First;
          nbLast   : Last;
          nbInsert : Insert;
          nbEdit   : Edit;
          nbCancel : Cancel;
          nbPost   : Post;
          nbRefresh: Refresh;
          nbDelete :
          IF NOT FConfirmDelete OR (MessageDlg('Ok to delete this record?',mtConfirmation,mbOKCancel,0) <> idCancel) THEN BEGIN
            Delete;
          END {IF};
        END {CASE};
      END {WITH};

      DataChanged;
      EditingChanged;
    END {IF};

    IF NOT (csDesigning IN ComponentState) AND Assigned(FOnNavClick) THEN BEGIN
      FOnNavClick(Self,aIndex);
    END {IF};
{}END {BtnClick};




{}PROCEDURE   TKNavigator.BtnMouseDown(aSender:TObject;aButton:TMouseButton;aShift:TShiftState;X,Y:INTEGER);
  VAR
    OldFocus  : TKNavBtn;
  BEGIN
    OldFocus := FFocusedButton;
    FFocusedButton := TKNavButton(aSender).INDEX;
    IF TabStop AND (GetFocus <> Handle) AND CanFocus THEN BEGIN
      SetFocus;
      IF (GetFocus <> Handle) THEN BEGIN
        Exit;
      END {IF};
    END ELSE
    IF TabStop AND (GetFocus = Handle) AND (OldFocus <> FFocusedButton) THEN BEGIN
      Buttons[OldFocus].Invalidate;
      Buttons[FFocusedButton].Invalidate;
    END {IF};
{}END {BtnMouseDown};




{}PROCEDURE   TKNavigator.ClickHandler(aSender:TObject);
  BEGIN
    BtnClick(TKNavButton(aSender).INDEX);
{}END {ClickHandler};




{}PROCEDURE   TKNavigator.CMEnabledChanged(VAR aMessage:TMessage);
  BEGIN
    INHERITED;
    IF NOT (csLoading IN ComponentState) THEN BEGIN
      ActiveChanged;
    END {IF};
{}END {CMEnabledChanged};




{}CONSTRUCTOR TKNavigator.Create(aOwner:TComponent);
  BEGIN
    INHERITED Create(aOwner);
    ControlStyle := ControlStyle-[csAcceptsControls,csSetCaption]+[csOpaque];
    IF NOT NewStyleControls THEN BEGIN
      ControlStyle := ControlStyle+[csFramed];
    END {IF};

    FVisibleButtons := [nbFirst,nbPrior,nbNext,nbLast,nbInsert,nbDelete,nbEdit,nbPost,nbCancel,nbRefresh,nbUser1,nbUser2];
    FHints := TStringList.Create;
    TStringList(FHints).OnChange := HintsChanged;
    InitButtons;
    BevelOuter := bvNone;
    BevelInner := bvNone;
    Width := 241;
    Height := 25;
    FButtonWidth := 0;
    FFocusedButton := nbFirst;
    FConfirmDelete := TRUE;
    FullRepaint := FALSE;
{}END {Create};




{}PROCEDURE   TKNavigator.DataChanged;
  VAR
    lfUpEnable: BOOLEAN;
    lfDnEnable: BOOLEAN;
    lfDeleteOk: BOOLEAN;
  BEGIN
    IF (FDataSet <> NIL) THEN BEGIN
      IF FdataSet.Active THEN BEGIN
        lfUpEnable := Enabled AND NOT DataSet.BOF;
        lfDnEnable := Enabled AND NOT DataSet.EOF;
        lfDeleteOk := Enabled AND DataSet.CanModify AND NOT (DataSet.BOF AND DataSet.EOF);
      END ELSE BEGIN
        lfUpEnable := FALSE;
        lfDnEnable := FALSE;
        lfDeleteOk := FALSE;
      END {IF};
    END ELSE BEGIN
      lfUpEnable := TRUE;
      lfDnEnable := TRUE;
      lfDeleteOk := TRUE;
    END {IF};

    Buttons[nbFirst].Enabled := lfUpEnable;
    Buttons[nbPrior].Enabled := lfUpEnable;
    Buttons[nbNext].Enabled  := lfDnEnable;
    Buttons[nbLast].Enabled  := lfDnEnable;
    Buttons[nbDelete].Enabled:= lfDeleteOk;
{}END {DataChanged};




{}PROCEDURE TKNavigator.DefineProperties(aFiler: TFiler);
  BEGIN
    INHERITED DefineProperties(aFiler);
    aFiler.DefineProperty('DataSetName', ReadDataSetName, WriteDataSetName, NOT StoreDataSet);
{}END {DefineProperties};




{}DESTRUCTOR  TKNavigator.Destroy;
  VAR
    leBtn     : TKNavBtn;
  BEGIN
    FHints.Free;
    FOR leBtn := Low(Buttons) TO High(Buttons) DO BEGIN
      Buttons[leBtn].Free;
    END {FOR};

    INHERITED Destroy;
{}END {Destroy};




{}PROCEDURE   TKNavigator.EditingChanged;
  VAR
    lfCanModify    : BOOLEAN;
    lfEditInsert   : BOOLEAN;
  BEGIN
    IF (FDataSet <> NIL) THEN BEGIN
      IF FDataSet.Active THEN BEGIN
        lfCanModify := Enabled AND FDataSet.CanModify;
        lfEditInsert := FDataSet.State IN [dsEdit,dsInsert];
      END ELSE BEGIN
        lfCanModify := FALSE;
        lfEditInsert := FALSE;
      END {IF};
    END ELSE BEGIN
      lfCanModify := Enabled;
      lfEditInsert := Enabled;
    END {IF};

    Buttons[nbInsert].Enabled := lfCanModify;
    Buttons[nbEdit].Enabled   := lfCanModify AND NOT lfEditInsert;
    Buttons[nbPost].Enabled   := lfCanModify AND lfEditInsert;
    Buttons[nbCancel].Enabled := lfCanModify AND lfEditInsert;
    Buttons[nbRefresh].Enabled:= lfCanModify;
    Buttons[nbUser1].Enabled  := Enabled;
    Buttons[nbUser2].Enabled  := Enabled;
{}END {EditingChanged};




{}PROCEDURE   TKNavigator.GetChildren(aProc:TGetChildProc;aRoot:TComponent);
  BEGIN
{}END {GetChildren};




{}FUNCTION    TKNavigator.GetDataSet:TDataSet;
  BEGIN
    Result := FDataSet;
{}END {GetDataSet};




{}PROCEDURE   TKNavigator.HintsChanged(aSender:TObject);
  BEGIN
    InitHints;
{}END {HintsChanged};




{}PROCEDURE   TKNavigator.InitButtons;
  VAR
    leBtn     : TKNavBtn;
    loBtn     : TKNavButton;
    liX       : INTEGER;
    liW,liH   : INTEGER;
  BEGIN
    FMinBtnSize := Point(20,18);
    liX := 0;
    FOR leBtn := Low(Buttons) TO High(Buttons) DO BEGIN
      loBtn := TKNavButton.Create(Self);
      loBtn.Flat := Flat;
      loBtn.Index := leBtn;
      loBtn.Visible := leBtn IN FVisibleButtons;
      loBtn.Enabled := TRUE;
      loBtn.SetBounds(liX,0,FMinBtnSize.X,FMinBtnSize.Y);
      IF FGlyphs <> NIL THEN BEGIN
        FGlyphs.GetBitMap(Ord(leBtn),loBtn.Glyph);
      END {IF};

      loBtn.NumGlyphs := 1;
      loBtn.Enabled := FALSE;
      loBtn.Enabled := TRUE;
      loBtn.OnClick := ClickHandler;
      loBtn.OnMouseDown := BtnMouseDown;
      loBtn.Parent := Self;
      IF Assigned(Buttons[leBtn]) THEN BEGIN
        Buttons[leBtn].Free;
      END {IF};

      Buttons[leBtn] := loBtn;
      liX := liX+FMinBtnSize.X;
    END {FOR};

    liW := Width;
    liH := Height;
    AdjustSize(liW,liH);
    IF (liW <> Width) OR (liH <> Height) THEN BEGIN
      INHERITED SetBounds(Left,Top,liW,liH);
    END {IF};

    InitHints;
    Buttons[nbPrior].NavStyle := Buttons[nbPrior].NavStyle+[nsAllowTimer];
    Buttons[nbNext].NavStyle := Buttons[nbNext].NavStyle+[nsAllowTimer];
{}END {InitButtons};




{}PROCEDURE   TKNavigator.InitHints;
  VAR
    li        : INTEGER;
    leBtn     : TKNavBtn;
  BEGIN
    leBtn := Low(Buttons);
    FOR li := 0 TO (FHints.Count-1) DO BEGIN
      IF FHints.Strings[li] <> '' THEN BEGIN
        Buttons[leBtn].Hint := FHints.Strings[li];
      END {IF};

      IF leBtn = High(Buttons) THEN BEGIN
        Exit;
      END {IF};

      Inc(leBtn);
    END {FOR};
{}END {InitHints};




{}PROCEDURE   TKNavigator.KeyDown(VAR aKey:WORD; aShift:TShiftState);
  VAR
    leNewFocus: TKNavBtn;
    leOldFocus: TKNavBtn;
  BEGIN
    leOldFocus := FFocusedButton;

    CASE aKey OF
      VK_RIGHT: BEGIN
        leNewFocus := FFocusedButton;
        REPEAT
          IF leNewFocus<High(Buttons) THEN BEGIN
            leNewFocus := Succ(leNewFocus);
          END {IF};
        UNTIL(leNewFocus = High(Buttons)) OR (Buttons[leNewFocus].Visible);

        IF leNewFocus <> FFocusedButton THEN BEGIN
          FFocusedButton := leNewFocus;
          Buttons[leOldFocus].Invalidate;
          Buttons[FFocusedButton].Invalidate;
        END {IF};
      END;

      VK_LEFT : BEGIN
        leNewFocus := FFocusedButton;
        REPEAT
          IF leNewFocus > Low(Buttons) THEN BEGIN
            leNewFocus := Pred(leNewFocus);
          END {IF};
        UNTIL(leNewFocus = Low(Buttons)) OR (Buttons[leNewFocus].Visible);

        IF leNewFocus <> FFocusedButton THEN BEGIN
          FFocusedButton := leNewFocus;
          Buttons[leOldFocus].Invalidate;
          Buttons[FFocusedButton].Invalidate;
        END {IF};
      END;

      VK_SPACE: BEGIN
        IF Buttons[FFocusedButton].Enabled THEN BEGIN
          Buttons[FFocusedButton].Click;
        END {IF};
      END;
    END {CASE};
{}END {KeyDown};




{}PROCEDURE   TKNavigator.Loaded;
  VAR
    liH       : INTEGER;
    liW       : INTEGER;
  BEGIN
    INHERITED Loaded;
    liW := Width;
    liH := Height;
    AdjustSize(liW,liH);
    IF (liW <> Width) OR (liH <> Height) THEN BEGIN
      INHERITED SetBounds(Left,Top,liW,liH);
    END {IF};

    InitHints;
    ActiveChanged;
{}END {Loaded};




{}PROCEDURE   TKNavigator.Notification(aComponent:TComponent; aOperation:TOperation);
  BEGIN
    INHERITED Notification(aComponent,aOperation);
    IF (aOperation = opRemove) THEN BEGIN
      IF (aComponent = DataSet) THEN BEGIN
        DataSet := NIL;
      END ELSE
      IF (aComponent = Glyphs) THEN BEGIN
        Glyphs := NIL;
      END {IF};
    END {IF};
{}END {Notification};




{}PROCEDURE   TKNavigator.ReadDataSetName(aReader : TReader);
  BEGIN
    SetDataSet(FindDataSet(aReader.ReadString));
{}END {ReadDataSetName};




{}PROCEDURE   TKNavigator.SetBounds(aLeft,aTop,aWidth,aHeight:INTEGER);
  VAR
    liW       : INTEGER;
    liH       : INTEGER;
  BEGIN
    liW := aWidth;
    liH := aHeight;
    IF NOT HandleAllocated THEN BEGIN
      AdjustSize(liW,liH);
    END {IF};

    INHERITED SetBounds(aLeft,aTop,liW,liH);
{}END {SetBounds};




{}PROCEDURE   TKNavigator.SetDataSet(aValue:TDataSet);
  BEGIN
    IF aValue <> FDataset THEN BEGIN
      FDataSet := aValue;
      ActiveChanged;
    END {IF};
{}END {SetDataSet};




{}PROCEDURE   TKNavigator.SetFlat(aValue:BOOLEAN);
  VAR
    leBtn     : TKNavBtn;
  BEGIN
    IF FFlat <> aValue THEN BEGIN
      FFlat := aValue;
      FOR leBtn := Low(Buttons) TO High(Buttons) DO BEGIN
        Buttons[leBtn].Flat := aValue;
      END {FOR};
    END {IF};
{}END {SetFlat};




{}PROCEDURE   TKNavigator.SetHints(aValue:TStrings);
  BEGIN
    FHints.Assign(aValue);
{}END {SetHints};




{}PROCEDURE   TKNavigator.SetImageList(aValue: TImageList);
  BEGIN
    FGlyphs := aValue;
    InitButtons;
{}END {SetImageList};




{}PROCEDURE   TKNavigator.SetVisible(aValue:TKNavBtnSet);
  VAR
    leBtn     : TKNavBtn;
    W,
    H         : INTEGER;
  BEGIN
    W := Width;
    H := Height;
    FVisibleButtons := aValue;
    FOR leBtn := Low(Buttons) TO High(Buttons) DO BEGIN
      Buttons[leBtn].Visible := leBtn IN FVisibleButtons;
    END {FOR};

    AdjustSize(W,H);
    IF (W <> Width) OR (H <> Height) THEN BEGIN
      INHERITED SetBounds(Left,Top,W,H);
    END {IF};

    Invalidate;
{}END {SetVisible};




{}FUNCTION    TKNavigator.StoreDataSet : BOOLEAN;
  BEGIN
    Result := (FDataSet = NIL) OR (Owner.FindComponent(FDataSet.Name) <> NIL);
{}END {StoreDataSet};




{}PROCEDURE   TKNavigator.WMGetDlgCode(VAR aMessage:TWMGetDlgCode);
{+M
---------------------------------------------------------------------------
  Purpose     - Notify Windows that we want to handle arrow keys.
---------------------------------------------------------------------------}
  BEGIN
    aMessage.Result := DLGC_WANTARROWS;
{}END {WMGetDlgCode};




{}PROCEDURE   TKNavigator.WMKillFocus(VAR aMessage:TWMKillFocus);
  BEGIN
    Buttons[FFocusedButton].Invalidate;
{}END {WMKillFocus};




{}PROCEDURE   TKNavigator.WMSetFocus(VAR aMessage:TWMSetFocus);
  BEGIN
    Buttons[FFocusedButton].Invalidate;
{}END {WMSetFocus};




{}PROCEDURE   TKNavigator.WMSize(VAR aMessage:TWMSize);
  VAR
    liW       : INTEGER;
    liH       : INTEGER;
  BEGIN
    INHERITED;
  { check for minimum size }
    liW := Width;
    liH := Height;
    AdjustSize(liW,liH);
    IF (liW <> Width) OR (liH <> Height) THEN BEGIN
      INHERITED SetBounds(Left,Top,liW,liH);
    END {IF};

    aMessage.Result := 0;
{}END {WMSize};




{}PROCEDURE   TKNavigator.WriteDataSetName(aWriter : TWriter);
  BEGIN
    IF FDataSet = NIL THEN
      aWriter.WriteString('')
    ELSE BEGIN
      aWriter.WriteString(FDataSet.Name);
    END {IF};
{}END {WriteDataSetName};


  {TKNavButton}


{}DESTRUCTOR  TKNavButton.Destroy;
  BEGIN
    IF FRepeatTimer <> NIL THEN BEGIN
      FRepeatTimer.Free;
    END {IF};

    INHERITED Destroy;
{}END {Destroy};




{}PROCEDURE   TKNavButton.MouseDown(aButton:TMouseButton;aShift:TShiftState;X,Y:INTEGER);
  BEGIN
    INHERITED MouseDown(aButton,aShift,X,Y);
    IF nsAllowTimer IN FNavStyle THEN BEGIN
      IF FRepeatTimer = NIL THEN BEGIN
        FRepeatTimer := TTimer.Create(Self);
      END {IF};

      FRepeatTimer.OnTimer := TimerExpired;
      FRepeatTimer.Interval := InitRepeatPause;
      FRepeatTimer.Enabled := TRUE;
    END {IF};
{}END {MouseDown};




{}PROCEDURE   TKNavButton.MouseUp(aButton:TMouseButton;aShift:TShiftState;X,Y:INTEGER);
  BEGIN
    INHERITED MouseUp(aButton,aShift,X,Y);
    IF FRepeatTimer <> NIL THEN BEGIN
      FRepeatTimer.Enabled := FALSE;
    END {IF};
{}END {MouseUp};




{}PROCEDURE   TKNavButton.Paint;
  VAR
    laR       : TRect;
  BEGIN
    INHERITED Paint;
    IF (GetFocus = Parent.Handle) AND (FIndex = TKNavigator(Parent).FFocusedButton) THEN BEGIN
      laR := Bounds(0,0,Width,Height);
      InflateRect(laR,-3,-3);
      IF FState = bsDown THEN BEGIN
        OffsetRect(laR,1,1);
      END {IF};

      DrawFocusRect(Canvas.Handle,laR);
    END {IF};
{}END {Paint};




{}PROCEDURE   TKNavButton.TimerExpired(aSender:TObject);
  BEGIN
    FRepeatTimer.Interval := RepeatPause;
    IF (FState = bsDown) AND MouseCapture THEN BEGIN
      TRY
        Click;
      EXCEPT
        FRepeatTimer.Enabled := FALSE;
        RAISE;
      END {TRY};
    END {IF};
{}END {TimerExpired};




{}FUNCTION    FindDataSet(CONST aValue:TComponentName):TDataSet;
{+M
---------------------------------------------------------------------------
 Purpose     - Returns the (pointer to the) dataset having the name aValue.
---------------------------------------------------------------------------}
  VAR
    liI       : INTEGER;
    loThis    : TComponent;
  BEGIN
    Result := NIL;
    IF aValue <> '' THEN BEGIN
      loThis := application.FindComponent(aValue);
      IF loThis <> NIL THEN BEGIN
        IF loThis IS TDataSet THEN BEGIN
          Result := (loThis AS TDataSet);
          Exit;
        END {IF};
      END ELSE BEGIN
        FOR liI := 0 TO application.ComponentCount-1 DO BEGIN
          IF (application.Components[liI] IS TForm) OR (application.Components[liI] IS TDataModule) THEN BEGIN
            loThis := application.Components[liI].FindComponent(aValue);
            IF loThis IS TDataSet THEN BEGIN
              Result := (loThis AS TDataSet);
              Exit;
            END {IF};
          END {IF};
        END {FOR};
      END {IF};
    END {IF};
{}END {FindDataSet};



END {KNavigat}.
