UNIT NKolCBox;
{+H
---------------------------------------------------------------------------
  File        - NKOLCBOX.PAS
  Purpose     - This component completes the text that a user writes in a
                combobox with items present on the list, however the list
                may contain multiple columns of data. Entry matching and
                completion is based on a specified column. The list box
                portion may be wider than the edit portion of the combo.
  Version     - TNKolComboBox; version 1.0 (1998.0305)
  Usage       - Component for Delphi 2.0, 3.0x and C++ Builder 1.0.
  Author      - Keith S. Brown
                E-mail: keith_brown@juno.com
                http://www.ghg.net/chairouchu/
                2437 Bay Area Blvd #20, Houston TX / USA
  Copyright    1998 Keith S. Brown
  Restrictions- This component is FREEWARE!
                No warranty is given by the author, expressed or implied.
                Do not distribute modified versions without my agreement.
  Revised     - 1998.03xx (KSB) Wrote initial version
              - 1998.0317 (KSB) Added sorting by reference column. Added
                SelectedColumnStr property.
              - 1998.0318 (KSB) Added IsRefColMember and CreateNewItem methods.
                Added PutColumnIntoString function.
              - 1998.0319 (KSB) Add Preview property.
---------------------------------------------------------------------------}
INTERFACE
USES
  Windows,  Messages, SysUtils, Classes,  Graphics, Controls, Forms,
  Dialogs,  StdCtrls, DsgnIntf;


CONST
  cMaxColumnNo= 256;
TYPE
  TColumnRange= 0..cMaxColumnNo-1;
  TColumnJust = (cjLeft, cjRight, cjCenter);
  TListStyle  = (psBoldStyle, psColumnStyle, psStandard);

  TColumnInfo = CLASS(TPersistent)
  PRIVATE
    FColumns       : TStringList;
    FOnChange      : TNotifyEvent;
  PROTECTED
    PROCEDURE   DefineProperties(aFiler: TFiler); OVERRIDE;
    FUNCTION    GetGap(aIndex:TColumnRange):INTEGER;
    FUNCTION    GetJust(aIndex:TColumnRange):TColumnJust;
    FUNCTION    GetNumColumns:INTEGER;
    FUNCTION    GetWidth(aIndex:TColumnRange):INTEGER;
    PROCEDURE   ReadColDefs(aReader: TReader); VIRTUAL;
    PROCEDURE   SetGap(aIndex:TColumnRange; aGap:INTEGER);
    PROCEDURE   SetJust(aIndex:TColumnRange; aJust:TColumnJust);
    PROCEDURE   SetWidth(aIndex:TColumnRange; aWidth:INTEGER);
    PROCEDURE   WriteColDefs(aWriter: TWriter); VIRTUAL;
  PUBLIC
    PROCEDURE   Assign(aSource: TPersistent); OVERRIDE;
    CLASS FUNCTION  ColJustToStr(aJust: TColumnJust): STRING;
    PROPERTY    Columns:TStringList READ FColumns;
    CONSTRUCTOR Create;
    CLASS PROCEDURE DecodeColStr(aColStr: STRING; VAR aWidth: INTEGER; VAR aJust: TColumnJust; VAR aGap: INTEGER);
    DESTRUCTOR  Destroy; OVERRIDE;
    CLASS FUNCTION  EncodeColStr(aWidth: INTEGER; aJust: TColumnJust; aGap: INTEGER): STRING;
    PROPERTY    Gap[aIndex: TColumnRange]: INTEGER READ GetGap WRITE SetGap;
    PROPERTY    Justify[aIndex: TColumnRange]: TColumnJust READ GetJust WRITE SetJust;
    PROPERTY    NumColumns: INTEGER READ GetNumColumns;
    CLASS FUNCTION  StrToColJust(aStr: STRING): TColumnJust;
    PROPERTY    Width[aIndex: TColumnRange]: INTEGER READ GetWidth WRITE SetWidth;
  PUBLISHED
    PROPERTY    OnChange: TNotifyEvent READ FOnChange WRITE FOnChange;
  END {CLASS};


  TCompareEvent    = PROCEDURE(aControl:TWinControl; CONST aStr1,aStr2:STRING; VAR aResult:INTEGER) OF OBJECT;
  TSortedListBox   = CLASS(TListBox)
  PRIVATE
    FfHorzScrollBar: BOOLEAN;     {T if horizontal scroll bar on}
    FwHorzExtent   : WORD;        {horizontal pixels for scroll bar}
    FOnCompare: TCompareEvent;
    FUNCTION    CompareStrings(CONST aStr1,aStr2:STRING):INTEGER;
  PROTECTED
    PROCEDURE   CMWantSpecialKey(VAR aMsg:TCMWantSpecialKey); MESSAGE CM_WANTSPECIALKEY;
    PROCEDURE   CreateParams(VAR aParams:TCreateParams); OVERRIDE;
    PROCEDURE   DoCreateWnd;
    PROCEDURE   SetOnCompare(aValue:TCompareEvent);
    PROCEDURE   WMGetDlgCode(VAR aMessage:TWMGetDlgCode); MESSAGE WM_GETDLGCODE;
    PROCEDURE   WMLButtonDown(VAR aMessage:TWMLButtonDown); MESSAGE WM_LBUTTONDOWN;
  PUBLIC
    PROCEDURE   DoSort;
  PUBLISHED
    PROPERTY    OnCompare:TCompareEvent READ FOnCompare WRITE SetOnCompare;
  END {CLASS};


  TCompleteEvent   = PROCEDURE(aSender:TObject; aKey:WORD; VAR aCompleteWith:STRING; VAR aDoComplete:BOOLEAN) OF OBJECT;
  TNKolumnComboBox = CLASS(TCustomEdit)
  PRIVATE
    { Private declarations for autocomplete}
    FfComplete     : BOOLEAN;     {T if autocomplete on}
    FsText         : STRING;      {initial edit text}
    FwKey          : WORD;
    FwMinLngth     : WORD;
    FOnComplete    : TCompleteEvent;
  PRIVATE
    { Other Private declarations }
    FeStyle        : TListStyle;
    FfListOnTop    : BOOLEAN;     {T is list is above editbox }
    FfPreview      : BOOLEAN;     {T to see listbox during design}
    FfSorted       : BOOLEAN;     {T if list is sorted}
    FiDropDownCount: INTEGER;     {maximum number of items to show in list box}
    FiItemHeight   : INTEGER;     {the height of one list box line}
    FiReferenceCol : TColumnRange;{when selected from the list, the value from this column is used, 0 is column 1}
    FiRefColColor  : TColor;      {color of reference column}
    FAbout         : TObject;
    FColumns       : TColumnInfo;      {column formatting info}
    FList          : TSortedListBox;   {list portion of combobox}
    FOnDropDown    : TNotifyEvent;

    PROCEDURE   CMFontChanged(VAR aMsg:TMessage); MESSAGE CM_FONTCHANGED;
    PROCEDURE   CMWantSpecialKey(VAR aMsg:TCMWantSpecialKey); MESSAGE CM_WANTSPECIALKEY;
    FUNCTION    GetDroppedDown:BOOLEAN;
    FUNCTION    GetItems:TStrings;
    FUNCTION    GetItemIndex: Integer;
    FUNCTION    GetListWidth: INTEGER;
    PROCEDURE   SetComplete(aValue:BOOLEAN);
    PROCEDURE   SetDroppedDown(aValue:BOOLEAN);
    PROCEDURE   SetDropDownCount(aValue:INTEGER);
    PROCEDURE   SetItems(aValue: TStrings);
    PROCEDURE   SetItemIndex(aValue: Integer);
    PROCEDURE   SetListWidth(aValue:INTEGER);
    PROCEDURE   SetMinLength(aValue:WORD);
    PROCEDURE   WMGetDlgCode(VAR aMessage:TWMGetDlgCode); MESSAGE WM_GETDLGCODE;
  PROTECTED
    { Protected declarations }
    PROCEDURE   Change; OVERRIDE;
    PROCEDURE   CreateWnd; OVERRIDE;
    PROCEDURE   DoEnter; OVERRIDE;
    PROCEDURE   DoExit; OVERRIDE;
    PROCEDURE   DropDown; DYNAMIC;
    FUNCTION    GetColumnStr(aColumn,aIndex:INTEGER):STRING;
    FUNCTION    GetSelectedColumnStr(aColumn:TColumnRange):STRING;
    FUNCTION    GetHorzScrollBar:BOOLEAN;
    FUNCTION    GetReferencedColStr(aIndex:INTEGER):STRING;
    PROCEDURE   KeyDown(VAR aKey:WORD; aShift:TShiftState); OVERRIDE;
    PROCEDURE   SetColumns(aColumns:TColumnInfo);
    PROCEDURE   SetHorzScrollBar(aValue:BOOLEAN);
    PROCEDURE   SetListOnTop(aValue:BOOLEAN);
    PROCEDURE   SetPreview(aValue:BOOLEAN);
    PROCEDURE   SetRefColColor(aValue:TColor);
    PROCEDURE   SetReferenceCol(aValue:TColumnRange);
    PROCEDURE   SetSorted(aValue:BOOLEAN);
    PROCEDURE   SetStyle(aValue:TListStyle);
    PROPERTY    StoredText:STRING READ FsText;
  PUBLIC
    { Public declarations }
    PROCEDURE   CompareRefCol(aSender:TWinControl; CONST aStr1,aStr2:STRING; VAR aResult:INTEGER); VIRTUAL;
    CONSTRUCTOR Create(aOwner : TComponent); OVERRIDE;
    DESTRUCTOR  Destroy; OVERRIDE;
    PROPERTY    DroppedDown:BOOLEAN READ GetDroppedDown WRITE SetDroppedDown;
    PROCEDURE   Clear; OVERRIDE;
    FUNCTION    CreateNewItem:STRING;
    FUNCTION    IsRefColMember(aStr:STRING):BOOLEAN;
    PROPERTY    ItemIndex: INTEGER READ GetItemIndex WRITE SetItemIndex;
    PROPERTY    SelectedColumnStr[aColumn:TColumnRange]:STRING READ GetSelectedColumnStr;
    PROCEDURE   SetBounds(aLeft,aTop,aWidth,aHeight:INTEGER); OVERRIDE;
  PUBLISHED
    { Published declarations }
    PROPERTY    About : TObject READ FAbout WRITE FAbout;
    PROPERTY    Color;
    PROPERTY    Columns: TColumnInfo READ FColumns WRITE SetColumns;
    PROPERTY    Complete:BOOLEAN READ FfComplete WRITE SetComplete;
    PROPERTY    Ctl3D;
    PROPERTY    Cursor;
    PROPERTY    DragCursor;
    PROPERTY    DropDownCount: INTEGER READ FiDropDownCount WRITE SetDropDownCount DEFAULT 8;
    PROPERTY    Enabled;
    PROPERTY    Font;
    PROPERTY    Height;
    PROPERTY    HelpContext;
    PROPERTY    Hint;
    PROPERTY    HorizScrollBar:BOOLEAN READ GetHorzScrollBar WRITE SetHorzScrollBar DEFAULT FALSE;
    PROPERTY    Items: TStrings READ GetItems WRITE SetItems;
    PROPERTY    Left;
    PROPERTY    ListOnTop: BOOLEAN READ FfListOnTop WRITE SetListOnTop DEFAULT FALSE;
    PROPERTY    ListWidth: INTEGER READ GetListWidth WRITE SetListWidth;
    PROPERTY    MaxLength;
    PROPERTY    MinLengthComplete:WORD READ FwMinLngth WRITE SetMinLength;
    PROPERTY    Name;
    PROPERTY    OnChange;
    PROPERTY    OnClick;
    PROPERTY    OnComplete:TCompleteEvent READ FOnComplete WRITE FOnComplete;
    PROPERTY    OnDblClick;
    PROPERTY    OnDragDrop;
    PROPERTY    OnDragOver;
    PROPERTY    OnDropDown: TNotifyEvent READ FOnDropDown WRITE FOnDropDown;
    PROPERTY    OnEndDrag;
    PROPERTY    OnEnter;
    PROPERTY    OnExit;
    PROPERTY    OnKeyDown;
    PROPERTY    OnKeyPress;
    PROPERTY    OnKeyUp;
    PROPERTY    ParentColor DEFAULT FALSE;
    PROPERTY    ParentCtl3D;
    PROPERTY    ParentFont;
    PROPERTY    ParentShowHint;
    PROPERTY    PopupMenu;
    PROPERTY    Preview: BOOLEAN READ FfPreview WRITE SetPreview;
    PROPERTY    ReferenceColumn:TColumnRange READ FiReferenceCol WRITE SetReferenceCol DEFAULT 0;
    PROPERTY    ReferenceColumnColor:TColor READ FiRefColColor WRITE SetRefColColor DEFAULT clBlack;
    PROPERTY    ShowHint;
    PROPERTY    Sorted:BOOLEAN READ FfSorted WRITE SetSorted DEFAULT TRUE;
    PROPERTY    Style:TListStyle READ FeStyle WRITE SetStyle;
    PROPERTY    TabOrder;
    PROPERTY    TabStop DEFAULT TRUE;
    PROPERTY    Tag;
    PROPERTY    Text;
    PROPERTY    Top;
    PROPERTY    Visible;
    PROPERTY    Width;
  END {CLASS};


{}FUNCTION    GetColumnFromString(CONST aStr:STRING; aColumn:TColumnRange):STRING;
{}FUNCTION    PutColumnIntoString(CONST aStr,aCol:STRING; aColumn:TColumnRange):STRING;
              {=================================================}
IMPLEMENTATION
{.$DEFINE DebugDrawColumnStyle}
{.$DEFINE DebugListBoxHeight}
{= = = = = = = = = = = = = =}
{$IFDEF DebugDrawColumnStyle}
  {$DEFINE DebugLogging}
{$ENDIF}
{$IFDEF DebugListBoxHeight}
  {$DEFINE DebugLogging}
{$ENDIF}
CONST
  chTAB       = #$09;
  chSPACE     = #$20;
{$IFDEF DebugLogging}
VAR
  uLOG        : TEXTFILE;
{$ENDIF}


{}FUNCTION    GetColumnFromString(CONST aStr:STRING; aColumn:TColumnRange):STRING;
  VAR
    li        : INTEGER;
    liPos     : INTEGER;
    liCol     : INTEGER;
    ls        : STRING;
  BEGIN
    ls := chTAB+aStr;
    liPos := 0;
    liCol := -1;
    FOR li := 1 TO Length(ls) DO BEGIN
      IF ls[li] = chTAB THEN BEGIN
        Inc(liCol);
        liPos := li;
        IF liCol = aColumn THEN BEGIN
          Break;
        END {IF};
      END {IF};
    END {FOR};

    IF liCol = aColumn THEN BEGIN
      Delete(ls,1,liPos);
      liPos := Pos(chTAB,ls);
      IF liPos = 0 THEN BEGIN
        result := ls;
      END ELSE BEGIN
        result := Copy(ls,1,liPos-1);
      END {IF};
    END ELSE BEGIN
      result := '';
    END {IF};
{}END {GetColumnFromString};




{}FUNCTION    PutColumnIntoString(CONST aStr,aCol:STRING; aColumn:TColumnRange):STRING;
  VAR
    li        : INTEGER;
    liPos     : INTEGER;
    liCol     : INTEGER;
    ls        : STRING;
  BEGIN
    result := chTAB+aStr;    {add leading tab}
    liPos := 0;
    liCol := -1;
    FOR li := 1 TO Length(result) DO BEGIN
      IF result[li] = chTAB THEN BEGIN
        Inc(liCol);
        liPos := li;
        IF liCol = aColumn THEN BEGIN
          Break;
        END {IF};
      END {IF};
    END {FOR};

    IF liCol = aColumn THEN BEGIN
      ls := Copy(result,liPos+1,Length(result)-liPos+1);
      result := Copy(result,1,liPos)+aCol;
      WHILE (Length(ls) > 0) AND (ls[1] <> chTAB) DO BEGIN
        Delete(ls,1,1);
      END {WHILE};

      result := result + ls;
    END ELSE BEGIN
      FOR li := liCol TO aColumn-1 DO BEGIN
        result := result+chTAB;
      END {FOR};

      result := result + aCol;
    END {IF};

    Delete(result,1,1);      {remove leading tab}
{}END {PutColumnIntoString};


{ TColumnInfo }


{}PROCEDURE   TColumnInfo.Assign(aSource: TPersistent);
  VAR
    li        : INTEGER;
  BEGIN
    IF (aSource IS TColumnInfo) THEN BEGIN
      IF (aSource = NIL) THEN BEGIN
        self.FColumns.Clear;
      END ELSE BEGIN
        IF aSource <> self THEN BEGIN
          self.FColumns.Clear;
          FOR li := 0 TO TColumnInfo(aSource).NumColumns-1 DO BEGIN
            self.FColumns.Add(TColumnInfo(aSource).FColumns[li]);
          END {FOR};
        END {IF};
      END {IF};

      IF Assigned(FOnChange) THEN BEGIN
        FOnChange(self);
      END {IF};
    END {IF};
{}END {Assign};




{}CONSTRUCTOR TColumnInfo.Create;
  BEGIN
    FColumns := TStringList.Create;
{}END {Create};




{}PROCEDURE   TColumnInfo.DefineProperties(aFiler: TFiler);
  BEGIN
    aFiler.DefineProperty('ColumnDefs', ReadColDefs, WriteColDefs, FColumns.Count > 0);
{}END {DefineProperties};




{}PROCEDURE   TColumnInfo.ReadColDefs(aReader: TReader);
  BEGIN
    aReader.ReadListBegin;
    FColumns.BeginUpdate;
    TRY
      FColumns.Clear;
      WHILE NOT aReader.EndOfList DO BEGIN
        FColumns.Add(aReader.ReadString);
      END {WHILE};
    FINALLY
      FColumns.EndUpdate;
    END {TRY};

    aReader.ReadListEnd;
{}END {ReadColDefs};




{}PROCEDURE   TColumnInfo.WriteColDefs(aWriter: TWriter);
  VAR
    li        : INTEGER;
  BEGIN
    aWriter.WriteListBegin;
    FOR li := 0 TO FColumns.Count - 1 DO BEGIN
      aWriter.WriteString(FColumns.Strings[li]);
    END {FOR};

    aWriter.WriteListEnd;
{}END {WriteColDefs};




{}DESTRUCTOR  TColumnInfo.Destroy;
  BEGIN
    FColumns.Free;
{}END {Destroy};




{}FUNCTION    TColumnInfo.GetGap(aIndex:TColumnRange):INTEGER;
  VAR
    liWidth   : INTEGER;
    leJust    : TColumnJust;
  BEGIN
    IF aIndex <= NumColumns-1 THEN BEGIN
      DecodeColStr(FColumns[aIndex],liWidth,leJust,result);
    END ELSE BEGIN
      result := 0;
    END {IF};
{}END {GetGap};




{}FUNCTION    TColumnInfo.GetJust(aIndex:TColumnRange):TColumnJust;
  VAR
    liGap     : INTEGER;
    liWidth   : INTEGER;
  BEGIN
    IF aIndex <= NumColumns-1 THEN BEGIN
      DecodeColStr(FColumns[aIndex],liWidth,result,liGap);
    END ELSE BEGIN
      result := cjLeft;
    END {IF};
{}END {GetJust};




{}FUNCTION    TColumnInfo.GetNumColumns:INTEGER;
  BEGIN
    result := FColumns.Count;
{}END {GetNumColumns};




{}FUNCTION    TColumnInfo.GetWidth(aIndex:TColumnRange):INTEGER;
  VAR
    liGap     : INTEGER;
    leJust    : TColumnJust;
  BEGIN
    IF aIndex <= NumColumns-1 THEN BEGIN
      DecodeColStr(FColumns[aIndex],result,leJust,liGap);
    END ELSE BEGIN
      result := 0;
    END {IF};
{}END {GetWidth};




{}PROCEDURE   TColumnInfo.SetGap(aIndex:TColumnRange; aGap:INTEGER);
  VAR
    liGap     : INTEGER;
    liWidth   : INTEGER;
    leJust    : TColumnJust;
  BEGIN
    IF aIndex <= NumColumns-1 THEN BEGIN
      DecodeColStr(FColumns[aIndex],liWidth,leJust,liGap);
      FColumns[aIndex] := EncodeColStr(liWidth,leJust,aGap);
    END ELSE
    IF aIndex < cMaxColumnNo THEN BEGIN
      FColumns.Add(EncodeColStr(10,cjLeft,aGap));
    END {IF};

    IF Assigned(FOnChange) THEN BEGIN
      FOnChange(self);
    END {IF};
{}END {SetGap};




{}PROCEDURE   TColumnInfo.SetJust(aIndex:TColumnRange; aJust:TColumnJust);
  VAR
    liGap     : INTEGER;
    liWidth   : INTEGER;
    leJust    : TColumnJust;
  BEGIN
    IF aIndex <= NumColumns-1 THEN BEGIN
      DecodeColStr(FColumns[aIndex],liWidth,leJust,liGap);
      FColumns[aIndex] := EncodeColStr(liWidth,aJust,liGap);
    END ELSE
    IF aIndex < cMaxColumnNo THEN BEGIN
      FColumns.Add(EncodeColStr(10,aJust,0));
    END {IF};

    IF Assigned(FOnChange) THEN BEGIN
      FOnChange(self);
    END {IF};
{}END {SetJust};




{}PROCEDURE   TColumnInfo.SetWidth(aIndex:TColumnRange; aWidth:INTEGER);
  VAR
    liGap     : INTEGER;
    liWidth   : INTEGER;
    leJust    : TColumnJust;
  BEGIN
    IF aIndex <= NumColumns-1 THEN BEGIN
      DecodeColStr(FColumns[aIndex],liWidth,leJust,liGap);
      FColumns[aIndex] := EncodeColStr(aWidth,leJust,liGap);
    END ELSE
    IF aIndex < cMaxColumnNo THEN BEGIN
      FColumns.Add(EncodeColStr(aWidth,cjLeft,0));
    END {IF};

    IF Assigned(FOnChange) THEN BEGIN
      FOnChange(self);
    END {IF};
{}END {SetWidth};




{}CLASS FUNCTION  TColumnInfo.ColJustToStr(aJust: TColumnJust): STRING;
  BEGIN
    CASE aJust OF
      cjLeft  : result := 'Left';
      cjRight : result := 'Right';
      cjCenter: result := 'Center';
    END {CASE};
{}END {ColJustToStr};




{}CLASS FUNCTION  TColumnInfo.EncodeColStr(aWidth: INTEGER; aJust: TColumnJust; aGap: INTEGER): STRING;
  BEGIN
    result := IntToStr(aWidth)+' '+ColJustToStr(aJust)[1]+' '+IntToStr(aGap);
{}END {EncodeColStr};




{}CLASS PROCEDURE TColumnInfo.DecodeColStr(aColStr: STRING; VAR aWidth: INTEGER; VAR aJust: TColumnJust; VAR aGap: INTEGER);
  CONST
    cWhiteSpace    = [' ',',',';'];
  TYPE
    ePortion  = (_width,_just,_gap);
  VAR
    ls        : STRING;
    le        : ePortion;
    li        : INTEGER;
    liStart   : INTEGER;
  BEGIN
    aWidth := 0;
    aJust  := cjLeft;
    aGap   := 0;
    aColStr:= Trim(aColStr) + ' ';
    le     := _width;
    li     := 1;
    liStart:= 1;
    WHILE li <= Length(aColStr) DO BEGIN
      IF aColStr[li] IN cWhiteSpace THEN BEGIN
        ls := Copy(aColStr,liStart,li-liStart);
        CASE le OF
          _width : BEGIN
            aWidth := StrToIntDef(ls,aWidth);
            le := _just;
          END;

          _just  : BEGIN
            aJust := StrToColJust(ls);
            le := _gap;
          END;

          _gap   : BEGIN
            aGap := StrToIntDef(ls,aGap);
            Exit;
          END;
        END {CASE};

        WHILE (li < Length(aColStr)) AND (aColStr[li] IN cWhiteSpace) DO BEGIN
          Inc(li);
        END {WHILE};

        liStart := li;
      END {IF};

      Inc(li);
    END {WHILE};
{}END {DecodeColStr};




{}CLASS FUNCTION  TColumnInfo.StrToColJust(aStr: STRING): TColumnJust;
  BEGIN
    IF Length(aStr) > 0 THEN BEGIN
      CASE UpCase(aStr[1]) OF
        'C' : result := cjCenter;
        'L' : result := cjLeft;
        'R' : result := cjRight;
        ELSE  result := cjLeft;
      END {CASE};
    END ELSE BEGIN
      result := cjLeft;
    END {IF};
{}END {StrToColJust};


  { TSortedListBox }


{}PROCEDURE   TSortedListBox.CMWantSpecialKey(VAR aMsg:TCMWantSpecialKey);
  BEGIN
    INHERITED;
    IF (CHAR(aMsg.CharCode) = chTAB) THEN BEGIN
      aMsg.Result := 1;
    END {IF};
{}END {CMWantSpecialKey};




{}FUNCTION    TSortedListBox.CompareStrings(CONST aStr1,aStr2:STRING):INTEGER;
  BEGIN
    IF Assigned(FOnCompare) THEN BEGIN
      FOnCompare(self,aStr1,aStr2,result);
    END ELSE BEGIN
      result := AnsiCompareStr(aStr1,aStr2);
    END {IF};
{}END {CompareStrings};




{}PROCEDURE   TSortedListBox.CreateParams(VAR aParams:TCreateParams);
  BEGIN
    INHERITED CreateParams(aParams);
    aParams.Style := aParams.Style OR lbs_UseTabStops OR ws_HScroll;
{}END {CreateParams};




{}PROCEDURE   TSortedListBox.DoCreateWnd;
  BEGIN
    IF FfHorzScrollBar THEN BEGIN
      Perform(lb_SetHorizontalExtent, FwHorzExtent,0);
      ShowScrollBar(handle,sb_Horz,FfHorzScrollBar);
    END ELSE BEGIN
      Perform(lb_SetHorizontalExtent, 1,0);
      ShowScrollBar(handle,sb_Horz,FALSE);
    END {IF};
{}END {DoCreateWnd};




{}PROCEDURE   TSortedListBox.DoSort;
  VAR
    li        : INTEGER;
    loSL      : TStringList;

{}{}FUNCTION FindWhere(CONST aStr:STRING):INTEGER;
{
  Purpose     - Binary search to find location of string in list
}
    VAR
      li      : INTEGER;
      liHi    : INTEGER;
      liLo    : INTEGER;
    BEGIN
      liLo := 0;
      liHi := loSL.Count-1;
      WHILE liLo <= liHi DO BEGIN
        li := (liLo+liHi) SHR 1;
        IF CompareStrings(loSL[li],aStr) < 0 THEN BEGIN
          liLo := Succ(li);
        END ELSE BEGIN
          liHi := Pred(li);
        END {IF};
      END {WHILE};

      result := liLo;
{}{}END {FindWhere};


  BEGIN
    Sorted := FALSE;
    loSL := TStringList.Create;
    TRY
      FOR li := 0 TO Items.Count-1 DO BEGIN
        loSL.InsertObject(FindWhere(Items[li]), Items[li], Items.Objects[li]);
      END {FOR};

      Items.Assign(loSL);
    FINALLY
      loSL.Free;
    END {TRY};
{}END {DoSort};




{}PROCEDURE   TSortedListBox.SetOnCompare(aValue:TCompareEvent);
  BEGIN
    FOnCompare := aValue;
    DoSort;
{}END {SetOnCompare};




{}PROCEDURE   TSortedListBox.WMGetDlgCode(VAR aMessage:TWMGetDlgCode);
  BEGIN
    INHERITED;
    aMessage.Result := aMessage.Result OR DLGC_WANTTAB;
{}END {WMGetDlgCode};




{}PROCEDURE   TSortedListBox.WMLButtonDown(VAR aMessage: TWMLButtonDown);
{+M
---------------------------------------------------------------------------
  Purpose     -
---------------------------------------------------------------------------}
  VAR
    lw        : WORD;
  BEGIN
    IF (DragMode=dmAutomatic) AND (aMessage.XPos < (Width - GetSystemMetrics(SM_CXHSCROLL))) THEN BEGIN
      SetFocus;
      BeginDrag(FALSE);
      Exit;
    END {IF};

    INHERITED;
    IF (DragMode=dmManual) AND (aMessage.Xpos < (Width - GetSystemMetrics(SM_CXHSCROLL))) THEN BEGIN
      lw := VK_RETURN;
      KeyUp(lw,[]);
    END {IF};
{}END {WMLButtonDown};




TYPE
  TCBParent   = CLASS(TWinControl)
  END {CLASS};

  TCBList     = CLASS(TSortedListBox)
  PUBLIC
//  FpWindowProc   : TWndMethod;
    FCompanion     : TNKolumnComboBox;
    PROCEDURE   AdjustHorizExtent;
    PROCEDURE   AdjustListBoxHeight;
    PROCEDURE   AdjustListBoxTop;
    PROCEDURE   Change;
    CONSTRUCTOR Create(aOwner:TComponent); OVERRIDE;
    DESTRUCTOR  Destroy; OVERRIDE;
    PROCEDURE   DoExit; OVERRIDE;
    PROCEDURE   DrawColumnStyle(aIndex:INTEGER; aRect:TRect; aState:TOwnerDrawState; aStyle:TListStyle);
    PROCEDURE   DrawIgnoreTabStyle(aIndex:INTEGER; aRect:TRect; aState:TOwnerDrawState);
    PROCEDURE   DrawItem(aIndex:INTEGER; aRect:TRect; aState:TOwnerDrawState); OVERRIDE;
    PROCEDURE   KeyDown(VAR aKey:WORD; aShift:TShiftState); OVERRIDE;
    PROCEDURE   KeyUp(VAR aKey:WORD; aShift:TShiftState); OVERRIDE;
//  PROCEDURE   MyWndProc(VAR aMessage: TMessage); 
  END {CLASS};



{}PROCEDURE   TCBList.AdjustHorizExtent;
  VAR
    laMetrics : TTextMetric;
    li        : INTEGER;
    lw        : WORD;
  BEGIN
    GetTextMetrics(Canvas.Handle, laMetrics);
    lw := 0;
    WITH FCompanion DO BEGIN
      FOR li := 0 TO Columns.NumColumns-1 DO BEGIN
        lw := lw + Columns.Width[li]*laMetrics.tmAveCharWidth + Columns.Gap[li];
      END {FOR};
    END {WITH};

    IF lw > 0 THEN BEGIN
      FwHorzExtent := lw;
    END {IF};
{}END {AdjustHorizExtent};




{}PROCEDURE   TCBList.AdjustListBoxHeight;
  VAR
    liItemCount   : INTEGER;
  BEGIN
    liItemCount := Items.Count;
    IF liItemCount > FCompanion.FiDropDownCount THEN BEGIN
      liItemCount := FCompanion.FiDropDownCount;
    END {IF};

    IF liItemCount < 1 THEN BEGIN
      liItemCount := 1;
    END {IF};

    { adjust for integralHeight }
    IF (csDesigning IN ComponentState) THEN BEGIN
      IF FCompanion.FfPreview THEN BEGIN
        Height := FCompanion.FiDropDownCount*FCompanion.FiItemHeight+GetSystemMetrics(sm_CyBorder)*4;
      END ELSE BEGIN
        Height := 0;
      END {IF};
    END ELSE BEGIN
      Height := liItemCount*FCompanion.FiItemHeight+GetSystemMetrics(sm_CyBorder)*4;
      IF Visible THEN BEGIN
{$IFDEF DebugListBoxHeight}
  WriteLn(uLOG,'AdjustListBoxHeight - Height:',height:3,' Left:',left:3,' Top:',top:3,' Width:',width);
{$ENDIF}
        BringToFront;
      END {IF};
    END {IF};
{}END {AdjustListBoxHeight};




{}PROCEDURE   TCBList.AdjustListBoxTop;
  BEGIN
    IF FCompanion.FfListOnTop THEN BEGIN
      Top := FCompanion.Top - Height;
    END ELSE BEGIN
      Top := FCompanion.Top + FCompanion.Height;
    END {IF};
{}END {AdjustListBoxTop};




{}PROCEDURE   TCBList.Change;
  BEGIN
    FCompanion.Text := FCompanion.GetReferencedColStr(ItemIndex);
{}END {Change};




{}CONSTRUCTOR TCBList.Create(aOwner:TComponent);
  BEGIN
    INHERITED Create(aOwner);
    Visible:= FALSE;
    IntegralHeight := TRUE;
    Style  := lbOwnerDrawFixed;
    Sorted := FALSE;
    BorderStyle := bsSingle;
    MultiSelect := FALSE;
{}END {Create};




{}DESTRUCTOR  TCBList.Destroy;
{+M
---------------------------------------------------------------------------
  Remarks     - This destructor includes a work-around.

  If the list box is visible (ie, it is the top-most Z-order window and the form's close
  button is clicked [that is the [X] button in the upper right corner] an external
  exception is generated. As near as I can figure, the exception has to do with the
  fact the list box is the top-most Z-ordered window.

  If some other control receives focus [and therefore the listbox loses its top-most
  status] no problem occurs when the [X] button is clicked.

  I've tried to capture the WM_NCLBUTTONDOWN message so that I can SendToBack before the
  listbox is destroyed. I have not been successful. The message is easily captured from
  the form, but apparently impossible to capture from inside this component (a
  TCBList.WndProc method is not called when the user clicks the [X] button).

  I've tried replacing the form's WndProc from inside this component with one that
  captures the desired message... (using the WindowProc property and a custom WndProc
  method) and consistently wind up with a stack overflow for my troubles.
  .......
  When the user clicks [X] and the list box is visible, TCBList.Destroy is called before
  the TNKolumnComboBox destructor. When that destructor calls FList.Free an external
  exception is generated.

  If the user clicks [X] and the list box is NOT visible, TNKolumnComboBox.Destroy is
  called, which then calls FList.Free and everything works as it should.

  The workaround is to set the FCompanion.FList value to NIL after the inherited
  destructor for the list box is called. The TNKolumnComboBox.Destroy method then only
  calls FList.Free if FList is not NIL.

  Better solutions are certainly welcome.

  KSB-1998.0317
---------------------------------------------------------------------------}
  BEGIN
    INHERITED;
    FCompanion.FList := NIL;
    FCompanion := NIL;
{}END {Destroy};




{}PROCEDURE   TCBList.DoExit;
  BEGIN
    INHERITED;
    IF NOT FCompanion.Focused THEN BEGIN
      SendToBack;
      Visible := FALSE;
    END {IF};
{}END {DoExit};




{}PROCEDURE   TCBList.DrawColumnStyle(aIndex:INTEGER; aRect:TRect; aState:TOwnerDrawState; aStyle:TListStyle);
  VAR
    laMetrics : TTextMetric;
    laRect    : TRect;
    li        : INTEGER;
    liCount   : INTEGER;
    liFntColor: TColor;
    liLastCol : INTEGER;
    liW       : INTEGER;
    liX       : INTEGER;
    ls        : STRING;
  BEGIN
{$IFDEF DebugDrawColumnStyle}
    WriteLn(uLOG,'DrawColumnStyle(aIndex:',aIndex,
                 '; aRect:[(',aRect.left,',',aRect.top,
                 '),(',aRect.right,',',aRect.bottom,
                 ')]; aState:TOwnerDrawState; aStyle:TListStyle);');
{$ENDIF}
    liCount := FCompanion.Columns.Columns.Count;
    IF liCount > 0 THEN BEGIN
      liLastCol := Pred(liCount);
      GetTextMetrics(Canvas.Handle, laMetrics);
      WITH Canvas DO BEGIN
        Brush.Color := self.Color;
        FillRect(aRect);
        {manually set these colors to override the color change when the item is focused}
        Pen.Color := self.Font.Color;
        liFntColor:= self.Font.Color;
        IF odSelected IN aState THEN BEGIN
          CASE aStyle OF
            psBoldStyle : BEGIN
              Brush.Color:= self.Color;
              Font.Style := Font.Style + [fsBold];
            END;

            psColumnStyle : BEGIN
              Brush.Color := clHighlight;
              liFntColor := clHighlightText
            END;
          END {CASE};
        END ELSE BEGIN
          CASE aStyle OF
            psBoldStyle  : Font.Style := Font.Style - [fsBold];
          END {CASE};
        END {IF};

        FillChar(laRect,SizeOf(laRect),0);
        laRect.Top := aRect.Top;
        laRect.Bottom:= aRect.Bottom;
        FOR li := 0 TO liLastCol DO BEGIN
          ls := FCompanion.GetColumnStr(li,aIndex);
          IF li = FCompanion.ReferenceColumn THEN BEGIN
            IF (odSelected IN aState) AND (aStyle = psColumnStyle) THEN BEGIN
              Font.Color := liFntColor;
            END ELSE BEGIN
              Font.Color := FCompanion.FiRefColColor;
            END {IF};
          END ELSE BEGIN
            Font.Color := liFntColor;
          END {IF};

          liW := FCompanion.Columns.Width[li] * laMetrics.tmAveCharWidth;
          liX := FCompanion.Columns.Gap[li];
          CASE FCompanion.Columns.Justify[li] OF
            cjRight : liX := liW - Canvas.TextWidth(ls) - liX;
            cjCenter: liX := (liW - Canvas.TextWidth(ls)) DIV 2;
          END {CASE};

          laRect.Left := laRect.Right;
          laRect.Right:= laRect.Right+liW;
{$IFDEF DebugDrawColumnStyle}
WITH laRect DO
WriteLn(uLOG,'Col:',li:2,' W:',liW:3,' X:',liX:3,' ClientWidth:',ClientWidth:3,' laRect:[(',left,',',top,'),(',right,',',bottom,')] S:',ls);
{$ENDIF}
          TextRect(laRect,laRect.Left+liX,laRect.Top,ls);
          IF li < liLastCol THEN BEGIN
            IF ColorToRGB(color) <> ColorToRGB(clBtnFace) THEN BEGIN
              Pen.Color := clBtnFace;
            END ELSE BEGIN
              Pen.Color := clBtnShadow;
            END {IF};

            Canvas.MoveTo(laRect.Right-1,laRect.Top);
            Canvas.LineTo(laRect.Right-1,laRect.Bottom);
            Inc(laRect.Right);
            IF laRect.Right >= aRect.Right{ClientWidth} THEN BEGIN
              Break;
            END {IF};
          END {IF};
        END {FOR};
      END {WITH};
    END ELSE BEGIN
      DrawIgnoreTabStyle(aIndex,aRect,aState);
    END {IF};
{}END {DrawColumnStyle};




{}PROCEDURE   TCBList.DrawIgnoreTabStyle(aIndex:INTEGER; aRect:TRect; aState:TOwnerDrawState);
  VAR
    laRect    : TRect;
    li        : INTEGER;
    liCount   : INTEGER;
    liFntColor: INTEGER;
    liLastCol : INTEGER;
    ls        : STRING;
  BEGIN
{$IFDEF DebugDrawColumnStyle}
    WriteLn(uLOG,'DrawIgnoreTabStyle(aIndex:',aIndex,
                 '; aRect:[(',aRect.left,',',aRect.top,
                 '),(',aRect.right,',',aRect.bottom,
                 ')]; aState:TOwnerDrawState);');
{$ENDIF}
    WITH Canvas DO BEGIN
      Pen.Color := self.Font.Color;
      IF odSelected IN aState THEN BEGIN
        Brush.Color := clHighlight;
        liFntColor := clHighlightText
      END ELSE BEGIN
        Brush.Color := self.Color;
        liFntColor := self.Font.Color;
      END {IF};

      FillRect(aRect);
      liCount := FCompanion.Columns.Columns.Count;
      IF liCount > 0 THEN BEGIN
        liLastCol := liCount-1;
        FillChar(laRect,SizeOf(laRect),0);
        laRect.Top := aRect.Top;
        laRect.Bottom := aRect.Bottom;
        FOR li := 0 TO liLastCol DO BEGIN
          ls := FCompanion.GetColumnStr(li,aIndex)+' ';
          IF li = FCompanion.ReferenceColumn THEN BEGIN
            Font.Color := FCompanion.FiRefColColor;
          END ELSE BEGIN
            Font.Color := liFntColor;
          END {IF};

          laRect.Left := laRect.Right;
          laRect.Right:= laRect.Right + Canvas.TextWidth(ls);
{$IFDEF DebugDrawColumnStyle}
WITH laRect DO
WriteLn(uLOG,'Col:',li:2,' ClientWidth:',ClientWidth:3,' laRect:[(',left,',',top,'),(',right,',',bottom,')] S:',ls);
{$ENDIF}
          TextRect(laRect,laRect.Left,laRect.Top,ls);
        END {FOR};
      END ELSE BEGIN
        ls := Items[aIndex];
        FOR li := 0 TO Length(ls) DO BEGIN
          IF ls[li] = chTAB THEN BEGIN
            ls[li] := chSPACE;
          END {IF};
        END {FOR};

        TextOut(aRect.Left + 2, aRect.Top, ls);
      END {IF};
    END {WITH};
{}END {DrawIgnoreTabStyle};




{}PROCEDURE   TCBList.DrawItem(aIndex:INTEGER; aRect:TRect; aState:TOwnerDrawState);
  BEGIN
    CASE FCompanion.Style OF
      psStandard: DrawIgnoreTabStyle(aIndex,aRect,aState); //INHERITED DrawItem(aIndex, aRect, aState);
      ELSE        DrawColumnStyle(aIndex, aRect, aState, FCompanion.Style);
    END {CASE};
{}END {DrawItem};




{}PROCEDURE   TCBList.KeyDown(VAR aKey:WORD; aShift:TShiftState);
  BEGIN
    CASE aKey OF
      VK_TAB : BEGIN
        FCompanion.SetFocus;
        FCompanion.KeyDown(aKey,aShift);
      END;
    END {CASE};

    INHERITED KeyDown(aKey,aShift);
{}END {KeyDown};




{}PROCEDURE   TCBList.KeyUp(VAR aKey:WORD; aShift:TShiftState);
  BEGIN
    CASE aKey OF
      VK_TAB : BEGIN
        FCompanion.SetFocus;
      END;

      VK_RETURN : BEGIN
        FCompanion.Text := FCompanion.GetReferencedColStr(ItemIndex);
        FCompanion.SetFocus;
      END;

      VK_ESCAPE : BEGIN
        FCompanion.Text := FCompanion.FsText;
        FCompanion.SetFocus;
      END;

      ELSE BEGIN
        Change;
      END;
    END {CASE};

    INHERITED KeyUp(aKey,aShift);
{}END {KeyUp};




(*PROCEDURE   TCBList.MyWndProc(VAR aMessage: TMessage);
  BEGIN
    CASE aMessage.Msg OF
      WM_NCLBUTTONDOWN : BEGIN
        IF Assigned(FpWindowProc) THEN BEGIN
          (self.Owner AS TControl).WindowProc := FpWindowProc;
          FpWindowProc := NIL;
          SendToBack;
          Exit;
        END {IF};
      END;

      ELSE BEGIN
        IF Assigned(FpWindowProc) THEN BEGIN
          FpWindowProc(aMessage);
        END ELSE BEGIN
          WndProc(aMessage);
        END {IF};
      END;
    END {CASE};
{}END {MyWndProc}; *)




{ TNKolumnComboBox }


{}PROCEDURE   TNKolumnComboBox.Change;
  VAR
    lfD       : BOOLEAN;
    lfDoIt    : BOOLEAN;
    lw        : WORD;
    lwD       : WORD;
    ls        : STRING;
  BEGIN
    INHERITED;
    IF (FfComplete) THEN BEGIN
      IF (Length(Text) >= FwMinLngth) AND (items.Count > 0) THEN BEGIN
        {don't trigger on a "TAB" or "ESCAPE" key}
        IF (FwKey<>VK_TAB) AND (FwKey<>VK_ESCAPE) THEN BEGIN
          ls := UpperCase(Text);            {prepare for case insensitve search}
          lfD:= (FwKey=VK_BACK) OR (FwKey=VK_DELETE);
          FOR lw := 0 TO items.Count-1 DO BEGIN
            IF ls = UpperCase(Copy(GetReferencedColStr(lw),1,Length(ls))) THEN BEGIN
              IF lfD THEN BEGIN             {"BACKSPACE" or "DELETE" key}
                ls := Text;
                ItemIndex := lw;
                Text := ls;
                SelStart := Length(Text);
                SelLength:= Length(Text);
              END ELSE BEGIN                {all keys, not [BACKSPACE,DELETE,TAB,ESCAPE]}
                lwD := Length(ls);
                ls := GetReferencedColStr(lw);
                ItemIndex := lw;
                lfDoIt := TRUE;
                IF Assigned(FOnComplete) THEN BEGIN
                  FOnComplete(Self,FwKey,ls,lfDoIt);
                END {IF};

                IF lfDoIt THEN BEGIN
                  Text := ls;
                  SelStart := lwD;
                  SelLength:= Length(Text);
                END {IF};
              END {IF};

              Break;
            END {IF};
          END {FOR};
        END {IF};
      END {IF};
    END {IF};
{}END {Change};




{}PROCEDURE   TNKolumnComboBox.Clear;
  BEGIN
    INHERITED Clear;
    FList.Clear;
{}END {Clear};




{}PROCEDURE   TNKolumnComboBox.CMFontChanged(VAR aMsg:TMessage);
  BEGIN
    INHERITED;
    TCBList(FList).Font := Font;
{}END {CMFontChanged};




{}PROCEDURE   TNKolumnComboBox.CMWantSpecialKey(VAR aMsg:TCMWantSpecialKey);
  BEGIN
    INHERITED;
    IF (CHAR(aMsg.CharCode) = #09) THEN BEGIN
      aMsg.Result := 1;
    END {IF};
{}END {CMWantSpecialKey};




{}PROCEDURE   TNKolumnComboBox.CompareRefCol(aSender:TWinControl; CONST aStr1,aStr2:STRING; VAR aResult:INTEGER);
  BEGIN
    aResult := AnsiCompareText(GetColumnFromString(aStr1,FiReferenceCol),GetColumnFromString(aStr2,FiReferenceCol));
{}END {CompareRefCol};




{}CONSTRUCTOR TNKolumnComboBox.Create(aOwner : Tcomponent);
  BEGIN
{$IFDEF DebugLogging}
    System.Assign(uLOG,ChangeFileExt(application.ExeName,'.LOG'));
    Rewrite(uLOG);
{$ENDIF}
    INHERITED Create(aOwner);
    Height := 25;
    Width := 145;
    FiItemHeight := 16;
    ParentColor := FALSE;
    TabStop := TRUE;

    ControlStyle := ControlStyle - [csAcceptsControls];
    ControlStyle := ControlStyle + [csCaptureMouse, csSetCaption, csDoubleClicks, csFixedHeight, csFramed];
    FfSorted := TRUE;
    FiReferenceCol := 0;
    FiRefColColor  := clBlack;
    FList := TCBList.Create(self);
    WITH FList AS TCBList DO BEGIN
      FCompanion := self;
      Width  := self.Width;
      Parent := TWinControl(aOwner);
    END {WITH};

    DropDownCount := 8;
    FeStyle := psColumnStyle;
    FColumns := TColumnInfo.Create;
    {autocomplete fields}
    FfComplete := TRUE;
    FwKey := 0;
    FwMinLngth := 0;
    FsText := '';
{}END {Create};




{}FUNCTION    TNKolumnComboBox.CreateNewItem:STRING;
  BEGIN
    result := PutColumnIntoString('','',FColumns.NumColumns-1);
    result := PutColumnIntoString(result,Text,ReferenceColumn);
{}END {CreateNewItem};




{}PROCEDURE   TNKolumnComboBox.CreateWnd;
  BEGIN
    INHERITED CreateWnd;
    (FList AS TCBList).AdjustListBoxTop;
    FList.Left:= Left;
    FList.DoCreateWnd;
{}END {CreateWnd};




{}DESTRUCTOR  TNKolumnComboBox.Destroy;
{+M
---------------------------------------------------------------------------
  Remarks     - See remarks in TCBList.Destroy
---------------------------------------------------------------------------}
  BEGIN
    FColumns.Free;
    IF Assigned(FList) THEN BEGIN
      FList.Free;
    END {IF};
    
    INHERITED Destroy;
{$IFDEF DebugLogging}
    CloseFile(uLOG);
{$ENDIF}    
{}END {Destroy};




{}PROCEDURE   TNKolumnComboBox.DoEnter;
  BEGIN
    INHERITED;
    DroppedDown := TRUE;
    IF FsText = '' THEN BEGIN
      FsText := Text;
      ItemIndex := -1;
    END {IF};
{}END {DoEnter};




{}PROCEDURE   TNKolumnComboBox.DoExit;
  BEGIN
    INHERITED;
    IF NOT FList.Focused THEN BEGIN
      DroppedDown := FALSE;
    END {IF};
{}END {DoExit};




{}PROCEDURE   TNKolumnComboBox.DropDown;
  BEGIN
    IF Assigned(FOnDropDown) THEN BEGIN
      FOnDropDown(Self);
    END {IF};
{}END {DropDown};




{}FUNCTION    TNKolumnComboBox.GetColumnStr(aColumn,aIndex:INTEGER):STRING;
  BEGIN
    TRY
      result := GetColumnFromString(Items[aIndex],aColumn);
    EXCEPT
      result := '';
    END {TRY};
{}END {GetColumnStr};




{}FUNCTION    TNKolumnComboBox.GetDroppedDown:BOOLEAN;
  BEGIN
    result := FList.visible;
{}END {GetDroppedDown};




{}FUNCTION    TNKolumnComboBox.GetHorzScrollBar:BOOLEAN;
  BEGIN
    result := FList.FfHorzScrollBar;
{}END {GetHorzScrollBar};




{}FUNCTION    TNKolumnComboBox.GetItemIndex:INTEGER;
  BEGIN
    result := FList.ItemIndex;
{}END {GetItemLndex};




{}FUNCTION    TNKolumnComboBox.GetItems:TStrings;
  BEGIN
    result := FList.Items;
{}END {GetItems};




{}FUNCTION    TNKolumnComboBox.GetListWidth:INTEGER;
  BEGIN
    result := FList.Width;
{}END {GetListWidth};




{}FUNCTION    TNKolumnComboBox.GetReferencedColStr(aIndex:INTEGER):STRING;
  BEGIN
    IF aIndex > -1 THEN BEGIN
      result := GetColumnStr(FiReferenceCol,aIndex);
    END ELSE BEGIN
      result := '';
    END {IF};
{}END {GetReferencedColStr};




{}FUNCTION    TNKolumnComboBox.GetSelectedColumnStr(aColumn:TColumnRange):STRING;
  BEGIN
    IF ItemIndex > -1 THEN BEGIN
      result := GetColumnStr(aColumn,ItemIndex);
    END ELSE BEGIN
      result := '';
    END {IF};
{}END {GetSelectedColumnStr};




{}FUNCTION    TNKolumnComboBox.IsRefColMember(aStr:STRING):BOOLEAN;
  VAR
    li        : INTEGER;
  BEGIN
    result := FALSE;
    aStr := UpperCase(aStr);
    FOR li := 0 TO FList.Items.Count-1 DO BEGIN
      IF aStr = UpperCase(GetReferencedColStr(li)) THEN BEGIN
        result := TRUE;
        Exit;
      END {IF};
    END {FOR};
{}END {IsRefColMember};




{}PROCEDURE   TNKolumnComboBox.KeyDown(VAR aKey:WORD; aShift:TShiftState);
  BEGIN
    INHERITED KeyDown(aKey,aShift);
    FwKey := aKey;
    CASE aKey OF
      VK_UP,VK_DOWN,
      VK_NEXT,VK_PRIOR : BEGIN
        FList.SetFocus;
        TCBList(FList).KeyDown(aKey,aShift);
      END;

      VK_TAB : BEGIN
        DroppedDown := FALSE;
        FsText := '';
        TCBParent(parent).SelectNext(self,NOT(ssCtrl IN aShift),TRUE);
      END;

      VK_ESCAPE : BEGIN
        ItemIndex := -1;
        Text := StoredText;
        SelStart := Length(Text);
        SelLength:= Length(Text);
      END;
    END {CASE};
{}END {KeyDown};




{}PROCEDURE   TNKolumnComboBox.SetBounds(aLeft,aTop,aWidth,aHeight : INTEGER );
  BEGIN
    INHERITED SetBounds(aLeft,aTop,aWidth,aHeight);
    IF Assigned(FList) THEN BEGIN
      FList.Left := Left;
      (FList AS TCBList).AdjustListBoxTop;
    END {IF};
{}END {SetBounds};




{}PROCEDURE   TNKolumnComboBox.SetColumns(aColumns:TColumnInfo);
  BEGIN
    FColumns.Assign(aColumns);
{}END {SetColumns};




{}PROCEDURE   TNKolumnComboBox.SetComplete(aValue:BOOLEAN);
  BEGIN
    FfComplete := aValue;
{}END {SetComplete};




{}PROCEDURE   TNKolumnComboBox.SetDropDownCount(aValue:INTEGER);
  BEGIN
    IF FiDropDownCount <> aValue THEN BEGIN
      FiDropDownCount := aValue;
      (FList AS TCBList).AdjustListBoxHeight;
      (FList AS TCBList).AdjustListBoxTop;
      FList.Invalidate;
    END {IF};
{}END {SetDropDownCount};




{}PROCEDURE   TNKolumnComboBox.SetDroppedDown(aValue:BOOLEAN);
{+M
---------------------------------------------------------------------------
  Remarks     - The Z-Order of the FList must be de-set BEFORE it is made
                invisible. Otherwise an external exception will arise when
                the application terminates.
---------------------------------------------------------------------------}
  BEGIN
    IF aValue <> FList.Visible THEN BEGIN
      IF aValue THEN BEGIN
        FList.Visible := TRUE;
        (FList AS TCBList).AdjustListBoxHeight;
        (FList AS TCBList).AdjustListBoxTop;
        (FList AS TCBList).AdjustHorizExtent;
        DropDown;
        FList.Invalidate;
      END ELSE BEGIN
        (FList AS TCBList).SendToBack;
        FList.visible := FALSE;
      END {IF};
    END {IF};
{}END {SetDroppedDown};




{}PROCEDURE   TNKolumnComboBox.SetHorzScrollBar(aValue:BOOLEAN);
  BEGIN
    IF aValue <> FList.FfHorzScrollBar THEN BEGIN
      FList.FfHorzScrollBar := aValue;
      FList.DoCreateWnd;
      FList.Invalidate;
    END {IF};
{}END {SetHorzScrollBar};




{}PROCEDURE   TNKolumnComboBox.SetItemIndex(aValue: INTEGER);
  BEGIN
    FList.ItemIndex := aValue;
{}END {SetItemIndex};




{}PROCEDURE   TNKolumnComboBox.SetItems(aValue:TStrings);
  VAR
    li        : INTEGER;
  BEGIN
    IF aValue <> FList.Items THEN BEGIN
      FList.Clear;
      FOR li := 0 TO aValue.Count-1 DO BEGIN
        FList.Items.Add(aValue[li]);
      END {FOR};

      (FList AS TCBList).AdjustListBoxHeight;
    END {IF};
{}END {SetItems};




{}PROCEDURE   TNKolumnComboBox.SetListOnTop(aValue:BOOLEAN);
  BEGIN
    IF aValue <> FfListOnTop THEN BEGIN
      FfListOnTop := aValue;
      (FList AS TCBList).AdjustListBoxTop;
      FList.Invalidate;
    END {IF};
{}END {SetListOnTop};




{}PROCEDURE   TNKolumnComboBox.SetListWidth(aValue:INTEGER);
  BEGIN
    FList.Width := aValue;
    (FList AS TCBList).AdjustHorizExtent;
{}END {SetListWidth};




{}PROCEDURE   TNKolumnComboBox.SetMinLength(aValue:WORD);
  BEGIN
    FwMinLngth := aValue;
{}END {SetMinLength};




{}PROCEDURE   TNKolumnComboBox.SetPreview(aValue:BOOLEAN);
  BEGIN
    IF aValue <> FfPreview THEN BEGIN
      FfPreview := aValue;
      (FList AS TCBList).AdjustListBoxHeight;
    END {IF};
{}END {SetPreview};




{}PROCEDURE   TNKolumnComboBox.SetRefColColor(aValue:TColor);
  BEGIN
    IF aValue <> FiRefColColor THEN BEGIN
      FiRefColColor := aValue;
      FList.Invalidate;
    END {IF};
{}END {SetRefColColor};




{}PROCEDURE   TNKolumnComboBox.SetReferenceCol(aValue:TColumnRange);
  BEGIN
    IF aValue <> FiReferenceCol THEN BEGIN
      FiReferenceCol := aValue;
      IF FfSorted THEN BEGIN
        FList.OnCompare := CompareRefCol;
      END ELSE BEGIN
        FList.OnCompare := NIL;
      END {IF};
      
      FList.Invalidate;
    END {IF};
{}END {SetReferenceCol};




{}PROCEDURE   TNKolumnComboBox.SetSorted(aValue:BOOLEAN);
  BEGIN
    IF aValue <> FfSorted THEN BEGIN
      FfSorted := aValue;
      IF FfSorted THEN BEGIN
        FList.OnCompare := CompareRefCol;
      END ELSE BEGIN
        FList.OnCompare := NIL;
      END {IF};

      FList.Invalidate;
    END {IF};
{}END {SetSorted};




{}PROCEDURE   TNKolumnComboBox.SetStyle(aValue:TListStyle);
  BEGIN
    IF aValue <> FeStyle THEN BEGIN
      FeStyle := aValue;
      FList.Invalidate;
    END {IF};
{}END {SetStyle};




{}PROCEDURE   TNKolumnComboBox.WMGetDlgCode(VAR aMessage:TWMGetDlgCode);
  BEGIN
    INHERITED;
    aMessage.Result := aMessage.Result OR DLGC_WANTTAB;
{}END {WMGetDlgCode};




END {NKolCBox}.
